#! /usr/bin/env python3

"""
pyRSKTools is a simple Python toolbox to open RSK SQLite files generated
by RBR instruments.
"""

from collections import namedtuple, OrderedDict
from datetime import datetime, timezone
import itertools
import re
import sqlite3

import ctdfjorder.pyrsktools.channel_types as channel_types

INSTRUMENT_TIME_MIN = datetime(2000, 1, 1, tzinfo=timezone.utc)
INSTRUMENT_TIME_MAX = datetime(2100, 1, 1, tzinfo=timezone.utc)

__copyright__ = "Copyright (c) 2017–2019 RBR Ltd"


class Channel(
    namedtuple("Channel", ["id", "key", "label", "name", "units", "derived"])
):
    __doc__ = """
Each physical sensor on an instrument is represented by one or more logical
channels.
"""

    def label(self):
        return "%s (%s)" % (self.name, self.units)


Deployment = namedtuple(
    "Deployment",
    [
        "id",
        "comment",
        "logger_status",
        "logger_time_drift",
        "download_time",
        "name",
        "sample_size",
    ],
)
Deployment.__doc__ = """
The RSK represents a deployment of the instrument.
"""

Instrument = namedtuple(
    "Instrument", ["serial", "model", "firmware_version", "firmware_type"]
)
Instrument.__doc__ = """
Instruments collect data.
"""

Geo = namedtuple(
    "Geo", ["timestamp", "latitude", "longitude", "accuracy", "accuracy_type"]
)
Geo.__doc__ = """
Geographic position data. May or may not be present in a dataset, depending on
the mode of collection.
"""


def auto_repr(cls):
    def __repr__(self):
        return "%s(%s)" % (
            type(self).__name__,
            ", ".join(
                "%s=%r" % var
                for var in vars(self).items()
                if not var[0].startswith("_")
            ),
        )

    cls.__repr__ = __repr__
    return cls


@auto_repr
class RSK(object):
    """
    An RSK dataset produced by Ruskin.
    """

    def __init__(self, rsk):
        # The deployment name gives the name of the dataset at time of download,
        # not the filename we're using to access it right now. Keep the current
        # name around in case someone needs it.
        self.name = rsk
        self._db = sqlite3.connect("file:%s?mode=ro" % rsk, uri=True)

        # Retrieve channels from the dataset.
        self.channels = OrderedDict()
        used_labels = {}
        for row in self._db.execute(
            """select channelID,
                                              shortName,
                                              longName,
                                              units,
                                              isDerived
                                         from channels
                                     order by channelId asc"""
        ):
            channel_id, key, name, units, derived = row
            derived = bool(derived)

            # RSKs don't store channel labels (and L2 channels don't have
            # labels naturally), so we'll generate them the same way Ruskin
            # otherwise would.
            label_prefix = channel_types.label_prefixes.get(
                key, name.lower().replace(" ", "")
            )
            label_counter = used_labels.get(label_prefix, 0)
            used_labels[label_prefix] = label_counter + 1
            label = "%s_%02d" % (label_prefix, label_counter)

            channel = Channel(channel_id, key, label, name, units, derived)
            self.channels[label] = channel

        # Mobile versions of Ruskin don't calculate derived channels, so RSKs
        # they generate won't have a full complement of data columns. We'll work
        # out which are present in the dataset we're dealing with so we can
        # offer the user convenient field names for samples.
        self.sample_channels = [
            column
            for column in _column_names(self._db, "data")
            if column.startswith("channel")
        ]
        self.sample_fields = ["timestamp"] + [
            channel
            for channel in self.channels
            if "channel%02d" % self.channels[channel].id in self.sample_channels
        ]
        self.Sample = namedtuple("Sample", self.sample_fields)

        raw_deployment = self._db.execute(
            """select deploymentID,
                                                    comment,
                                                    loggerStatus,
                                                    loggerTimeDrift,
                                                    timeOfDownload,
                                                    name,
                                                    sampleSize
                                               from deployments
                                              limit 1"""
        ).fetchone()
        logger_time_drift = raw_deployment[3] if raw_deployment[3] is not None else 0
        download_time = datetime.fromtimestamp(
            raw_deployment[4] / 1000, tz=timezone.utc
        )
        self.deployment = Deployment(
            *raw_deployment[:3]
            + (logger_time_drift, download_time)
            + raw_deployment[5:]
        )

        # Some older RSK versions kept the `firmwareVersion` and `firmwareType`
        # fields on the `deployments` table; newer versions keep them in the
        # `instruments` table. We'll sneakily use a join to pull the field from
        # wherever it exists. We don't need to join _on_ anything because there
        # should only be one row in each table (and the `serialID` field may
        # not exist on `deployments`).
        instrument = self._db.execute(
            """select instruments.serialID,
                                                model,
                                                firmwareVersion
                                           from instruments, deployments
                                          limit 1"""
        ).fetchone()

        # Much older RSKs don't have `firmwareType` at all, so we'll go looking
        # for that separately.
        if "firmwareType" in itertools.chain(
            _column_names(self._db, "instruments"),
            _column_names(self._db, "deployments"),
        ):
            firmware_type = self._db.execute(
                """select firmwareType
                                                  from instruments, deployments
                                                 limit 1"""
            )
            instrument = itertools.chain(instrument, firmware_type)
        else:
            instrument = itertools.chain(instrument, (None,))

        self.instrument = Instrument(*instrument)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def geodata(self, start_time=INSTRUMENT_TIME_MIN, end_time=INSTRUMENT_TIME_MAX):
        """
        Retrieve geographic information
        """
        for row in self._query_geodata(start_time, end_time):
            yield Geo(*row)

    def samples(self, start_time=INSTRUMENT_TIME_MIN, end_time=INSTRUMENT_TIME_MAX):
        """
        Retrieve samples from the dataset.
        """
        for row in self._query_samples(start_time, end_time):
            yield self.Sample(*row)

    def npsamples(self, start_time=INSTRUMENT_TIME_MIN, end_time=INSTRUMENT_TIME_MAX):
        """
        Retrieve samples from the dataset into a NumPy array. Requires NumPy.

        Unlike the “samples” method, which returns a generator, this method
        loads all requested data into memory. Be careful when working with large
        datasets! Rather than loading the entire dataset into memory, things may
        go more smoothly if you work with limited time ranges or a single
        cast/profile at a time.
        """
        try:
            import numpy as np
        except ImportError:
            np = None
            pass

        if not np:
            print("npsamples requires numpy, which could not be imported!")
            return

        # NumPy arrays are fixed-size, so first we need to figure out how many
        # samples we're retrieving.
        sample_count = self._db.execute(
            """select count(*)
                                             from data
                                            where tstamp >= ?
                                              and tstamp < ?""",
            (int(start_time.timestamp() * 1000), int(end_time.timestamp() * 1000)),
        ).fetchone()[0]

        # Allocate some space to put the samples...
        samples = np.zeros(
            sample_count,
            dtype={
                "names": self.sample_fields,
                "formats": ["object"] + ["float64"] * (len(self.sample_fields) - 1),
            },
        )

        # ...then load them in.
        i = 0
        for row in self._query_samples(start_time, end_time):
            samples[i] = row
            i += 1

        return samples

    def _query_geodata(self, start_time, end_time):
        return self._query(
            "geodata",
            itertools.chain(("tstamp",), _field_to_column_names(Geo._fields[1:])),
            start_time,
            end_time,
        )

    def _query_samples(self, start_time, end_time):
        return self._query(
            "data",
            itertools.chain(("tstamp",), self.sample_channels),
            start_time,
            end_time,
        )

    def _query(self, table, columns, start_time, end_time):
        if not table in _table_names(self._db):
            return

        column_list = ", ".join(columns)
        start_time = int(start_time.timestamp() * 1000)
        end_time = int(end_time.timestamp() * 1000)

        for row in self._db.execute(
            """select %s
                                         from %s
                                        where tstamp >= ?
                                          and tstamp < ?"""
            % (column_list, table),
            (start_time, end_time),
        ):
            timestamp = datetime.fromtimestamp(row[0] / 1000, tz=timezone.utc)
            yield (timestamp,) + row[1:]

    def profiles(self):
        """
        Retrieve the endpoints of all profiles in the dataset.
        """
        query = """select region.*
                     from region
                    where type = 'PROFILE'"""
        params = ()
        return self._query_regions(query, params)

    def casts(self, direction):
        """
        Retrieve the endpoints of directional casts in the dataset. Argument
        should be one of `pyrsktools.Region.CAST_DOWN`
        or `pyrsktools.Region.CAST_UP`.
        """
        query = """select region.*
                     from regionCast, region
                    where region.regionID = regionCast.regionID
                      and regionCast.type = ?"""
        params = (direction,)
        return self._query_regions(query, params)

    def _query_regions(self, query, params):
        if "region" not in _table_names(self._db):
            return

        cur = self._db.cursor()
        cur.execute(query, params)
        names = {
            name[0]: index
            for name, index in zip(cur.description, range(len(cur.description)))
        }

        for row in self._db.execute(query, params):
            start_time = datetime.fromtimestamp(
                row[names["tstamp1"]] / 1000, tz=timezone.utc
            )
            end_time = datetime.fromtimestamp(
                row[names["tstamp2"]] / 1000, tz=timezone.utc
            )
            label = row[names["label"]] if "label" in names else None
            description = row[names["description"]] if "description" in names else None
            yield Region(self, start_time, end_time, label, description)

        cur.close()

    def close(self):
        self._db.close()


@auto_repr
class Region(object):
    """
    An arbitrary region of time in a dataset.
    """

    CAST_DOWN = "DOWN"
    CAST_UP = "UP"

    def __init__(self, rsk, start_time, end_time, label, description):
        self._rsk = rsk
        self.start_time = start_time
        self.end_time = end_time
        self.label = label
        self.description = description

    def samples(self):
        """
        Retrieve samples from within the region.
        """
        return self._rsk.samples(self.start_time, self.end_time)

    def npsamples(self):
        """
        Retrieve samples from within the region into a Numpy array.
        """
        return self._rsk.npsamples(self.start_time, self.end_time)


def open(rsk):
    """
    Open an RSK file. Returns an RSK object ready for consumption.
    """
    return RSK(rsk)


def _table_names(db):
    """
    Retrieve a list of table names from a database.
    """

    return [
        row[0]
        for row in db.execute(
            "select name from sqlite_master where type = 'table'"
        ).fetchall()
    ]


def _column_names(db, table):
    """
    Retrieve the column names of a database table.
    """

    # Can't use bound values for table names. String substitution it is.
    return [row[1] for row in db.execute("pragma table_info('%s')" % table).fetchall()]


def _field_to_column_names(names):
    """
    Convert field names to database column names.

    The convention of this project is to use snake_case for field names (in the
    belief that this is the most Pythonic approach); conflictingly, the
    convention of the RSK schema is to use camelCase. This function converts
    from snake case to camel case with the intent of being used to help
    auto-generate SQL queries.
    """
    return [re.sub(r"_(.)", lambda m: m.group(1).upper(), name) for name in names]



import os.path
from datetime import datetime
import getpass

from ctdfjorder.constants.constants import *
import polars as pl
import sys
from os import path, getcwd


def save_to_csv(data: pl.DataFrame, output_file: str, null_value: str | None):
    """
    Renames the columns of the CTD data table based on a predefined mapping and saves the
    data to the specified CSV file.

    Parameters
    ----------
    data : pl.DataFrame
        The CTD data table.
    output_file : str
        The output CSV file path.
    null_value : str
        The value that will fill blank cells in the data.
    """

    def relabel_ctd_data(label: str) -> str:
        data_label_mapping = {
            TIMESTAMP_LABEL: EXPORT_TIMESTAMP_LABEL,
            YEAR_LABEL: EXPORT_YEAR_LABEL,
            MONTH_LABEL: EXPORT_MONTH_LABEL,
            TEMPERATURE_LABEL: EXPORT_TEMPERATURE_LABEL,
            PRESSURE_LABEL: EXPORT_PRESSURE_LABEL,
            CHLOROPHYLL_LABEL: EXPORT_CHLOROPHYLL_LABEL,
            SEA_PRESSURE_LABEL: EXPORT_SEA_PRESSURE_LABEL,
            DEPTH_LABEL: EXPORT_DEPTH_LABEL,
            SALINITY_LABEL: EXPORT_SALINITY_LABEL,
            SPEED_OF_SOUND_LABEL: EXPORT_SPEED_OF_SOUND_LABEL,
            SPECIFIC_CONDUCTIVITY_LABEL: EXPORT_SPECIFIC_CONDUCTIVITY_LABEL,
            CONDUCTIVITY_LABEL: EXPORT_CONDUCTIVITY_LABEL,
            DENSITY_LABEL: EXPORT_DENSITY_LABEL,
            POTENTIAL_DENSITY_LABEL: EXPORT_POTENTIAL_DENSITY_LABEL,
            SALINITY_ABS_LABEL: EXPORT_SALINITY_ABS_LABEL,
            SURFACE_DENSITY_LABEL: EXPORT_SURFACE_DENSITY_LABEL,
            SURFACE_SALINITY_LABEL: EXPORT_SURFACE_SALINITY_LABEL,
            SURFACE_TEMPERATURE_LABEL: EXPORT_SURFACE_TEMPERATURE_LABEL,
            MELTWATER_FRACTION_LABEL: EXPORT_MELTWATER_FRACTION_LABEL,
            LONGITUDE_LABEL: EXPORT_LONGITUDE_LABEL,
            LATITUDE_LABEL: EXPORT_LATITUDE_LABEL,
            FILENAME_LABEL: EXPORT_FILENAME_LABEL,
            PROFILE_ID_LABEL: EXPORT_PROFILE_ID_LABEL,
            UNIQUE_ID_LABEL: EXPORT_UNIQUE_ID_LABEL,
            BV_LABEL: EXPORT_BV_LABEL,
            P_MID_LABEL: EXPORT_P_MID_LABEL,
            SECCHI_DEPTH_LABEL: EXPORT_SECCHI_DEPTH_LABEL,
            SITE_NAME_LABEL: EXPORT_SITE_NAME_LABEL,
            SITE_ID_LABEL: EXPORT_SITE_ID_LABEL
        }
        return data_label_mapping.get(label, label)

    # Rename columns
    renamed_data = data.rename(relabel_ctd_data)

    # Define the desired column order based on the mapping values
    ordered_columns = [
        EXPORT_FILENAME_LABEL,
        EXPORT_PROFILE_ID_LABEL,
        EXPORT_UNIQUE_ID_LABEL,
        EXPORT_SITE_NAME_LABEL,
        EXPORT_SITE_ID_LABEL,
        EXPORT_LONGITUDE_LABEL,
        EXPORT_LATITUDE_LABEL,
        EXPORT_TIMESTAMP_LABEL,
        EXPORT_YEAR_LABEL,
        EXPORT_MONTH_LABEL,
        EXPORT_TEMPERATURE_LABEL,
        EXPORT_PRESSURE_LABEL,
        EXPORT_DEPTH_LABEL,
        EXPORT_SEA_PRESSURE_LABEL,
        EXPORT_CHLOROPHYLL_LABEL,
        EXPORT_SALINITY_LABEL,
        EXPORT_SPECIFIC_CONDUCTIVITY_LABEL,
        EXPORT_CONDUCTIVITY_LABEL,
        EXPORT_DENSITY_LABEL,
        EXPORT_POTENTIAL_DENSITY_LABEL,
        EXPORT_SALINITY_ABS_LABEL,
        EXPORT_SURFACE_DENSITY_LABEL,
        EXPORT_SPEED_OF_SOUND_LABEL,
        EXPORT_SURFACE_SALINITY_LABEL,
        EXPORT_SURFACE_TEMPERATURE_LABEL,
        EXPORT_MELTWATER_FRACTION_LABEL,
        EXPORT_BV_LABEL,
        EXPORT_P_MID_LABEL,
        EXPORT_SECCHI_DEPTH_LABEL,
    ]

    # Reorder columns if they are present in the DataFrame
    present_columns = [col for col in ordered_columns if col in renamed_data.columns]
    reordered_data = renamed_data.select(present_columns)

    # Append any missing columns that were not in the specified order
    missing_columns = [
        col for col in renamed_data.columns if col not in present_columns
    ]
    if missing_columns:
        missing_data = renamed_data.select(missing_columns)
        reordered_data = pl.concat([reordered_data, missing_data], how="horizontal")

    # Create metadata
    if output_file == "ctdfjorder_data.csv":
        creation_date = datetime.now().strftime('%Y%m%d%H%M%S')
        user = getpass.getuser()
        metadata = f"_{creation_date}_{user}.csv"
        file = os.path.splitext(output_file)[0]
        reordered_data.write_csv(file+metadata, null_value=null_value)
    else:
        reordered_data.write_csv(output_file, null_value=null_value)
    return reordered_data


def get_cwd():
    """
    Gets the current working directory.

    Returns
    -------
    str
        The current working directory.
    """
    # Determine if application is a script file or frozen exe
    if getattr(sys, "frozen", False):
        working_directory_path = path.dirname(sys.executable)
    elif __file__:
        working_directory_path = getcwd()
    else:
        working_directory_path = getcwd()
    return working_directory_path



import polars as pl
from ctdfjorder.pyrsktools import RSK, Geo, Region
from ctdfjorder.constants.constants import *
from ctdfjorder.exceptions.exceptions import CTDError
import numpy as np
from os import path
from typing import Any

# Column label mapping from rsk to internal
rskLabels_to_labelInternal: dict[str, str] = {
    "temperature_00": TEMPERATURE_LABEL,
    "chlorophyll_00": CHLOROPHYLL_LABEL,
    "seapressure_00": SEA_PRESSURE_LABEL,
    "depth_00": DEPTH_LABEL,
    "salinity_00": SALINITY_LABEL,
    "speedofsound_00": SPEED_OF_SOUND_LABEL,
    "specificconductivity_00": SPECIFIC_CONDUCTIVITY_LABEL,
    "conductivity_00": CONDUCTIVITY_LABEL,
    "pressure_00": PRESSURE_LABEL,
}


def load_file_rsk(rbr_file_path: str = None) -> pl.DataFrame:
    """
    Loads and processes an RSK file, extracting profiles and adding geospatial information.

    Parameters
    ----------
    rbr_file_path : str, optional
        The file path to the RSK file.

    Returns
    -------
    pl.DataFrame
        The processed RSK file data.

    Raises
    ------
    CTDError
        If the RSK profile is empty or if no samples are found in the file.
    """


    def process_rsk(
            rsk_profile: pl.DataFrame,
            geo: [Geo, Any, Any] = None,
            filename: str = None,
    ) -> pl.DataFrame | None:
        """
        Processes an RSK profile dataframe, adding geospatial information.

        Parameters
        ----------
        rsk_profile : pl.DataFrame
            The RSK profile data to process.
        geo : Geo, optional
            Geospatial information generator.
        filename : str, optional
            The filename of the RSK profile.

        Returns
        -------
        pl.DataFrame | None
            The processed RSK profile dataframe with latitude and longitude columns added, or None if the profile is empty.

        Raises
        ------
        CTDError
            If the RSK profile is empty.
        """
        rsk_profile = rsk_profile.with_columns(
            pl.col(TIMESTAMP_LABEL)
            .cast(pl.String)
            .str.to_datetime(
                format="%Y-%m-%d %H:%M:%S%.f",
                time_zone=TIME_ZONE,
                time_unit=TIME_UNIT,
            )
            .cast(pl.Datetime(time_unit=TIME_UNIT))
            .dt.convert_time_zone(time_zone=TIME_ZONE)
        )
        if rsk_profile.is_empty():
            raise CTDError(filename=filename, message=ERROR_NO_SAMPLES)
        data = rsk_profile
        try:
            profile_geodata = next(geo)
            return data.with_columns(
                pl.lit(profile_geodata.latitude).alias(LATITUDE_LABEL),
                pl.lit(profile_geodata.longitude).alias(LONGITUDE_LABEL),
            )

        # No geodata found in rsk file
        except StopIteration:
            return data.with_columns(
                pl.lit(None).alias(LATITUDE_LABEL), pl.lit(None).alias(LONGITUDE_LABEL)
            )

    data = None
    filename = path.basename(rbr_file_path)
    rsk = RSK(rbr_file_path)
    num_profiles = 0
    # Processing for rsk files with cast regions, loops through casts and indicates separate casts with profile num
    rsk_casts_down = rsk.casts(Region.CAST_DOWN)
    for i, endpoints in enumerate(rsk_casts_down):
        rsk_numpy_array = np.array(
            rsk.npsamples(endpoints.start_time, endpoints.end_time)
        )
        for x, timestamp in enumerate(rsk_numpy_array[TIMESTAMP_LABEL]):
            rsk_numpy_array[TIMESTAMP_LABEL][x] = timestamp.strftime(TIME_FORMAT)
        profile_to_process = (
            pl.DataFrame(rsk_numpy_array)
            .rename(rskLabels_to_labelInternal)
            .drop_nulls()
            .with_columns(
                pl.lit(filename).alias(FILENAME_LABEL),
                pl.lit(num_profiles).alias(PROFILE_ID_LABEL),
            )
        )
        geodata = rsk.geodata(endpoints.start_time, endpoints.end_time)
        processed_profile = process_rsk(
            rsk_profile=profile_to_process, geo=geodata, filename=filename
        )
        if data is None:
            data = processed_profile
            num_profiles += 1
        elif processed_profile is not None:
            data = pl.concat([processed_profile, data], how=CONCAT_HOW)
            num_profiles += 1

    # Processing for rsk files without cast regions
    if type(data) is type(None) or data.is_empty():
        num_profiles = 0
        rsk_numpy_array = np.array(rsk.npsamples())
        for x, timestamp in enumerate(rsk_numpy_array[TIMESTAMP_LABEL]):
            rsk_numpy_array[TIMESTAMP_LABEL][x] = timestamp.strftime(TIME_FORMAT)
        profile = (
            pl.DataFrame(rsk_numpy_array)
            .rename(rskLabels_to_labelInternal)
            .with_columns(
                pl.lit(filename).alias(FILENAME_LABEL),
                pl.lit(num_profiles).alias(PROFILE_ID_LABEL),
            )
        )
        geodata = rsk.geodata()
        processed_profile = process_rsk(
            rsk_profile=profile, geo=geodata, filename=filename
        )
        if data is None:
            data = processed_profile
            num_profiles += 1
        elif processed_profile is not None:
            data = pl.concat([processed_profile, data], how=CONCAT_HOW)
            num_profiles += 1
        else:
            CTDError(message=ERROR_NO_SAMPLES, filename=filename)
    return data



import logging
import shutil
import sys
from argparse import ArgumentParser
from concurrent.futures import ProcessPoolExecutor, as_completed
from contextlib import ExitStack
from os import path, listdir, remove, mkdir, getcwd
from warnings import catch_warnings
import signal

import polars as pl
import psutil
from polars.exceptions import ChronoFormatWarning
from rich.console import Console
from rich.panel import Panel
from rich.pretty import Pretty
from rich.table import Table, box
from rich.status import Status
from rich.prompt import Confirm
from rich import print as richprint
from rich_argparse import RichHelpFormatter

from ctdfjorder.visualize import ctd_plot
from ctdfjorder.exceptions.exceptions import CTDError, Critical
from ctdfjorder.CTD import CTD
from ctdfjorder.utils.utils import save_to_csv
from ctdfjorder.constants.constants import *
from ctdfjorder.metadata.master_sheet import MasterSheet

console = Console(color_system="windows")


def process_ctd_file(
        file, plot, cached_master_sheet, master_sheet_path, verbosity, plots_folder, filters
):
    """
    Processes a CTD file through a series of data cleaning and analysis steps.

    Parameters
    ----------
    file : str
        The path to the CTD file.
    plot : bool
        Flag indicating whether to generate plots.
    cached_master_sheet : MasterSheet
        A cached instance of the master sheet for cross-checking site names.
    master_sheet_path : str
        The path to the master sheet.
    verbosity : int
        The verbosity level for logging.
    plots_folder : str
        The folder to save plots in.
    filters : list
        List of filters to apply to the data.

    Returns
    -------
    pl.DataFrame | None
        The processed data as a Polars DataFrame or None if an error occurred.
    list[str]
        The status of each processing step, represented by color codes ("green", "yellow", "red").

    Notes
    -----
    The function goes through several steps to process the CTD file, including filtering,
    removing upcasts and non-positive samples, cleaning salinity data, adding surface measurements,
    absolute salinity, density, potential density, mixed layer depth (MLD), and Brunt-Väisälä frequency squared (BF Squared).
    If plotting is enabled, it generates plots for potential density and salinity versus depth.
    """
    logger = setup_logging(verbosity)
    steps = [
        (
            "Load File",
            lambda: CTD(
                file,
                plot=plot,
                cached_master_sheet=cached_master_sheet,
                master_sheet_path=master_sheet_path,
            ),
        ),
        ("Filter", lambda data: data.filter_columns_by_range(filters=filters)),
        ("Remove Upcasts", lambda data: data.remove_upcasts()),
        ("Remove Negative", lambda data: data.remove_non_positive_samples()),
        (
            "Remove Invalid Salinity Values",
            lambda data: data.remove_invalid_salinity_values(),
        ),
        ("Clean Salinity AI", lambda data: data.clean("clean_salinity_ai")),
        (
            "Add Surface Measurements",
            lambda data: data.add_surface_salinity_temp_meltwater(),
        ),
        ("Add Absolute Salinity", lambda data: data.add_absolute_salinity()),
        ("Add Density", lambda data: data.add_density()),
        ("Add Potential Density", lambda data: data.add_potential_density()),
        ("Add MLD", lambda data: data.add_mld(20, "potential_density_avg")),
        ("Add BF Squared", lambda data: data.add_bf_squared()),
        ("Plot", lambda data: plot_data(data.get_df(), plots_folder)),
        ("Exit", lambda data: data.get_df()),
    ]

    status = []
    data = None
    with catch_warnings(record=True, action="once") as warning_list:
        warning_list_length = len(warning_list)
        for step_name, step_function in steps:
            try:
                if step_name == "Load File":
                    data = step_function()
                elif step_name == "Exit":
                    return step_function(data), status + ["green"]
                else:
                    step_function(data)
                status.append(
                    "yellow" if len(warning_list) > warning_list_length
                                and warning_list[-1].category is not ChronoFormatWarning else "green"
                )
                warning_list_length = len(warning_list)
            except CTDError as error:
                logger.error(error)
                status.extend(["red"] * (len(steps) - len(status)))
                return None, status
            except Exception as e:
                print(e)
                logger.exception(e)
                status.extend(["red"] * (len(steps) - len(status)))
                return None, status
    return None, status


def plot_data(my_data, plots_folder):
    """
    Generates plots for the given CTD data.

    Parameters
    ----------
    my_data : pl.DataFrame
        The CTD data to plot.
    plots_folder : str
        The folder to save plots in.

    Notes
    -----
    This function generates two types of plots: depth vs potential density and depth vs salinity.
    The plots are saved in the specified folder.
    """
    ctd_plot.plot_depth_vs(my_data, POTENTIAL_DENSITY_LABEL, plot_folder=plots_folder)
    ctd_plot.plot_depth_vs(my_data, SALINITY_LABEL, plot_folder=plots_folder)


def generate_status_table(status_table):
    """
    Generates a status table displaying the processing status for each file.

    Parameters
    ----------
    status_table : list[tuple]
        A list of tuples containing the file name and its processing status.

    Returns
    -------
    rich.table.Table
        A rich Table object displaying the status of each file.

    Notes
    -----
    The table includes columns for each processing step and uses color codes to indicate the status:
    green for success, yellow for warnings, and red for errors.
    """
    steps = [
        "Load File",
        "Filter",
        "Remove Upcasts",
        "Remove Negative Values",
        "Remove Invalid Salinity Values",
        "Clean Salinity Data",
        "Add Surface Measurements",
        "Add Absolute Salinity",
        "Add Density",
        "Add Potential Density",
        "Add MLD",
        "Add BF Squared",
        "Plot",
        "Exit",
    ]
    table = Table(box=box.SQUARE)
    table.add_column("File", width=30)
    for step in steps:
        table.add_column(step)
    for file, status in status_table:
        table.add_row(file, *[f"[{color}]•[/{color}]" for color in status])
    return table


def run_default(
        plot,
        master_sheet_path,
        max_workers,
        verbosity,
        output_file,
        debug_run,
        status_show,
        mapbox_access_token,
        filters,
):
    """
    Runs the default processing pipeline for CTD files.

    Parameters
    ----------
    plot : bool
        Flag indicating whether to generate plots.
    master_sheet_path : str
        The path to the master sheet.
    max_workers : int
        The maximum number of worker processes to use.
    verbosity : int
        The verbosity level for logging.
    output_file : str
        The path to the output CSV file.
    debug_run : bool
        Flag indicating whether to run in debug mode (processes only 20 files).
    status_show : bool
        Flag indicating whether to show the processing status.
    mapbox_access_token : str
        The Mapbox access token for generating interactive maps.
    filters : list or None
        List of filters to apply to the data.

    Notes
    -----
    This function orchestrates the entire CTD file processing pipeline. It sets up logging,
    processes the master sheet, and uses a process pool to handle multiple CTD files concurrently.
    The processing status of each file is tracked, and results are compiled and saved to a CSV file.
    If plotting is enabled, it generates plots and an interactive map.
    """
    logger = setup_logging(verbosity)
    status_table, results = [], []
    cached_master_sheet = None
    if not status_show:
        Console(quiet=True)

    plots_folder = path.join(get_cwd(), "ctdplots")
    files = get_ctd_filenames_in_dir(get_cwd(), [".rsk", ".csv"])[
            : 20 if debug_run else None
            ]
    total_files = len(files)
    remaining_files = total_files

    if master_sheet_path in files:
        total_files -= 1
        files.remove(master_sheet_path)

    if not files:
        raise CTDError(message="No '.rsk' or '.csv' found in this folder", filename="")

    with Status(
            f"Processing master sheet, this might take awhile",
            spinner="earth",
            console=console,
    ) as status_master_sheet:
        try:
            status_master_sheet.start()
            cached_master_sheet = (
                MasterSheet(master_sheet_path, with_crosschecked_site_names=False)
                if master_sheet_path
                else None
            )
        except Critical as e:
            status_master_sheet.stop()
            console.print(e, style="white on red")
            continue_no_mastersheet = Confirm.ask("Continue without mastersheet?")
            if not continue_no_mastersheet:
                sys.exit()

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        status_spinner_processing = Status(f"Processing {total_files} files. Press CTRL+Z to shutdown.",
                                           spinner="earth",
                                           console=console,
                                           )
        status_spinner_processing.start()
        try:
            futures = {
                executor.submit(
                    process_ctd_file,
                    file,
                    plot,
                    cached_master_sheet,
                    master_sheet_path,
                    verbosity,
                    plots_folder,
                    filters,
                ): file
                for file in files
            }
            for future in as_completed(futures):
                file = futures[future]
                result, status = future.result()
                if isinstance(result, pl.DataFrame):
                    results.append(result)
                    remaining_files -= 1
                else:
                    remaining_files -= 1
                if status_show:
                    status_table.append((path.basename(file), status))
                    status_spinner_processing.update(
                        status=f"Processing {remaining_files} files"
                    )
            status_spinner_processing.stop()
            if status_show:
                console.print(generate_status_table(status_table))
            status_spinner_shutdown = Status(
                "End of processing, terminating open profile pipelines",
                spinner_style="red",
            )
            status_spinner_shutdown.start()
            executor.shutdown(wait=True, cancel_futures=True)
            status_spinner_shutdown.stop()
            process_results(
                results,
                total_files,
                output_file,
                plot,
                plots_folder,
                mapbox_access_token,
            )

        except KeyboardInterrupt:
            status_spinner_processing.stop()
            status_spinner_shutdown = Status(
                "Shutdown message received, terminating open profile pipelines",
                spinner_style="red",
            )
            status_spinner_shutdown.start()
            executor.shutdown(wait=True, cancel_futures=True)
            status_spinner_shutdown.stop()
            for proc in psutil.process_iter():
                if proc.name == "Python":
                    proc.kill()


def process_results(
        results, total_files, output_file, plot, plots_folder, mapbox_access_token
):
    """
    Processes the results of the CTD file processing pipeline.

    Parameters
    ----------
    results : list[pl.DataFrame]
        The list of processed CTD data.
    total_files : int
        The total number of files processed.
    output_file : str
        The path to the output CSV file.
    plot : bool
        Flag indicating whether to generate plots.
    plots_folder : str
        The folder to save plots in.
    mapbox_access_token : str
        The Mapbox access token for generating interactive maps.

    Notes
    -----
    This function combines the processed data from all files, generates a summary of the data,
    and saves the combined data to a CSV file. If mapbox access token is provided, it generates various plots
    and an interactive map using Mapbox.
    """
    with console.screen():
        with Status(
                "Combining CTD profiles", spinner="earth", console=console
        ) as status_spinner_combining:
            df = pl.concat(results, how="diagonal")
            panel = Panel(
                Pretty(df.select("pressure", "salinity", "temperature").describe()),
                title="Overall stats",
                subtitle=f"Errors/Total: {total_files - len(results)}/{total_files}",
            )
            pl.Config.set_tbl_rows(-1)
            df_test = df.unique(subset=["filename", PROFILE_ID_LABEL], keep="first").select(
                pl.col("filename"), pl.col("unique_id"), pl.col(TIMESTAMP_LABEL), pl.col(PROFILE_ID_LABEL)
            )
            df_test.write_csv(
                path.join(get_cwd(), "UniqueIDs")
            )
            richprint(panel)
            df = save_to_csv(df, output_file, None)

        if mapbox_access_token:
            plot_results(df, mapbox_access_token)


def plot_results(df, mapbox_access_token):
    """
    Generates plots for the processed CTD data and an interactive map view.

    Parameters
    ----------
    df : pl.DataFrame
        The processed CTD data.
    mapbox_access_token : str
        The Mapbox access token for generating interactive maps.

    Notes
    -----
    This function generates an interactive map to visualize the data.
    """
    with Status(
            "Running interactive map view. To shutdown press CTRL+Z.",
            spinner="earth",
            console=console,
    ) as status_spinner_map_view:
        if mapbox_access_token:
            try:
                ctd_plot.plot_map(df, mapbox_access_token)
            except KeyboardInterrupt:
                for proc in psutil.process_iter():
                    if proc.name == "Python":
                        proc.kill()


def get_ctd_filenames_in_dir(directory, types):
    """
    Gets the filenames of CTD files in a directory.

    Parameters
    ----------
    directory : str
        The directory to search for CTD files.
    types : list[str]
        A list of file extensions to search for.

    Returns
    -------
    list[str]
        A list of CTD file paths in the directory.

    Notes
    -----
    This function searches the specified directory for files with the given extensions (e.g., .rsk, .csv)
    and returns a list of matching file paths.
    """
    return [
        path.join(directory, f)
        for f in listdir(directory)
        if any(f.endswith(t) for t in types)
    ]


def get_cwd():
    """
    Gets the current working directory.

    Returns
    -------
    str
        The current working directory.

    Notes
    -----
    If the script is running in a frozen state (e.g., packaged with PyInstaller), it returns the directory
    of the executable. Otherwise, it returns the standard current working directory.
    """
    return path.dirname(sys.executable) if getattr(sys, "frozen", False) else getcwd()


def reset_file_environment():
    """
    Resets the file environment by removing existing output files and creating necessary directories.

    Notes
    -----
    This function removes old output files and directories to ensure a clean environment for running
    the CTD file processing pipeline. It creates a new directory for plots.
    """
    cwd = get_cwd()
    for filename in [DEFAULT_OUTPUT_FILE, "ctdfjorder.log"]:
        file_path = path.join(cwd, filename)
        if path.isfile(file_path):
            remove(file_path)
    if path.isdir(path.join(get_cwd(), "ctdplots")):
        shutil.rmtree(path.join(get_cwd(), "ctdplots"))
    mkdir(path.join(get_cwd(), "ctdplots"))


def setup_logging(verbosity):
    """
    Sets up logging for the application.

    Parameters
    ----------
    verbosity : int
        The verbosity level for logging.

    Returns
    -------
    logging.Logger
        The configured logger.

    Notes
    -----
    This function configures the logging settings for the application, including setting the log level
    based on verbosity, formatting log messages, and handling signals for graceful termination.
    """
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)
    level = max(30 - (verbosity * 10), 10)
    for logger_name in [
        "tensorflow",
        "matplotlib",
        "sklearn",
        "werkzeug",
        "dash",
        "flask",
    ]:
        logging.getLogger(logger_name).setLevel(logging.ERROR)
    logger = logging.getLogger("ctdfjorder")
    logger.handlers.clear()
    formatter = logging.Formatter(
        "%(asctime)s, %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s", "%H:%M:%S"
    )
    file_log = logging.FileHandler(path.join(get_cwd(), "ctdfjorder.log"))
    logging.basicConfig(level=level)
    file_log.setFormatter(formatter)
    file_log.setLevel(level)
    logger.addHandler(file_log)
    return logger


def signal_handler(signal_received, frame):
    r"""
    Handles system signals to gracefully terminate the application.

    Parameters
    ----------
    signal_received : Any
        The received signal.
    frame : frame
        The current stack frame.

    Notes
    -----
    This function handles system signals such as SIGINT and SIGTERM to allow for graceful
    termination of the application, cleaning up resources as needed.
    """
    if signal_received == signal.SIGINT:
        return
    raise KeyboardInterrupt


def build_parser():
    """
    Builds the argument parser for the command-line interface.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser.

    Notes
    -----
    This function sets up the argument parser for the command-line interface, defining the available
    commands and their respective options. It uses RichHelpFormatterPlus for enhanced help text formatting.
    """
    parser = ArgumentParser(
        description="CTDFjorder", formatter_class=RichHelpFormatter
    )
    subparsers = parser.add_subparsers(dest="command", required=True)
    parser_default = subparsers.add_parser(
        "default",
        help="Run the default processing pipeline",
        formatter_class=RichHelpFormatter,
    )
    parser_fjord = subparsers.add_parser(
        "fjord",
        help="Run the Fjord Phyto processing pipeline",
    )
    parser_fjord.add_argument(
        "-t",
        "--token",
        type=str,
        default=None,
        help="MapBox token to enable interactive map plot",
    )
    add_arguments(parser_default)
    return parser


def build_parser_docs():
    """
    Builds the argument parser for the documentation.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser.

    Notes
    -----
    This function sets up the argument parser specifically for generating documentation, defining
    the available commands and their respective options without the enhanced formatting.
    """
    parser = ArgumentParser(
        description="CTDFjorder"
    )
    subparsers = parser.add_subparsers(dest="command", required=True)
    parser_default = subparsers.add_parser(
        "default",
        help="Run the default processing pipeline",
    )
    parser_fjord = subparsers.add_parser(
        "fjord",
        help="Run the Fjord Phyto processing pipeline",
    )
    parser_fjord.add_argument(
        "-t",
        "--token",
        type=str,
        default=None,
        help="MapBox token to enable interactive map plot",
    )
    add_arguments(parser_default)
    return parser


def add_arguments(parser):
    """
    Adds arguments to the argument parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The argument parser to add arguments to.

    Notes
    -----
    This function adds various command-line arguments to the parser, including options for plotting,
    verbosity, resetting the file environment, showing processing status, running in debug mode,
    specifying the master sheet path, setting the number of worker processes, providing a Mapbox token,
    and defining filters for data columns.
    """
    parser.add_argument("-p", "--plot", action="store_true", help="Generate plots")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbosity",
        default=0,
        help="Verbose logger output to ctdfjorder.log (repeat for increased verbosity)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=-1,
        default=0,
        dest="verbosity",
        help="Quiet output (show errors only)",
    )
    parser.add_argument(
        "-r", "--reset", action="store_true", help="Reset file environment"
    )
    parser.add_argument(
        "-s",
        "--show-status",
        action="store_true",
        help="Show processing status and pipeline status",
    )
    parser.add_argument(
        "-d", "--debug-run", action="store_true", help="Run 20 files for testing"
    )
    parser.add_argument("-m", "--mastersheet", type=str, help="Path to mastersheet")
    parser.add_argument(
        "-w", "--workers", type=int, nargs="?", const=1, help="Max workers"
    )
    parser.add_argument(
        "--token",
        type=str,
        default=None,
        help="MapBox token to enable interactive map plot",
    )
    parser.add_argument(
        "-o", "--output", type=str, default=str(DEFAULT_OUTPUT_FILE), help="Output file path"
    )
    parser.add_argument(
        "--filtercolumns",
        nargs="*",
        type=str,
        required=False,
        default=None,
        help="List of columns to filter",
        choices=LIST_LABELS,
    )
    parser.add_argument(
        "--filterupper",
        nargs="*",
        type=float,
        required=False,
        default=None,
        help="Upper bounds for the filtered columns",
    )
    parser.add_argument(
        "--filterlower",
        nargs="*",
        type=float,
        required=False,
        default=None,
        help="Lower bounds for the filtered columns",
    )


def main():
    """
    The main entry point for the application. Parses arguments and runs the appropriate processing pipeline.

    Notes
    -----
    This function initializes signal handlers, builds the argument parser, and parses the command-line arguments.
    Depending on the specified command, it runs the default or Fjord Phyto processing pipeline.
    """
    for sig in [signal.SIGINT, signal.SIGTSTP, signal.SIGTERM]:
        signal.signal(sig, signal_handler)
    parser = build_parser()
    args = parser.parse_args()
    if args.command == "default":
        filters = create_filters(args)
        if args.reset:
            reset_file_environment()
        display_config(args)
        run_default(
            plot=args.plot,
            master_sheet_path=args.mastersheet,
            max_workers=args.workers,
            verbosity=args.verbosity,
            output_file=args.output,
            debug_run=args.debug_run,
            status_show=args.show_status,
            mapbox_access_token=args.token,
            filters=filters,
        )
        sys.exit()
    if args.command == "fjord":
        reset_file_environment()
        display_config(args)
        run_default(
            plot=True,
            master_sheet_path="mastersheet.csv",
            max_workers=8,
            verbosity=3,
            output_file=DEFAULT_OUTPUT_FILE,
            debug_run=False,
            status_show=True,
            mapbox_access_token=args.token,
            filters=None,
        )


def create_filters(args):
    """
    Creates filters from command-line arguments.

    Parameters
    ----------
    args : argparse.Namespace
        The parsed command-line arguments.

    Returns
    -------
    list[tuple] | None
        A list of filters or None if no filters are specified.

    Notes
    -----
    This function generates a list of filters based on the specified columns, upper bounds, and lower bounds
    provided via command-line arguments. If not all required filter arguments are provided, it returns None.
    """
    if all(
            arg is not None
            for arg in [args.filtercolumns, args.filterupper, args.filterlower]
    ):
        return zip(args.filtercolumns, args.filterupper, args.filterlower)
    return None


def display_config(args):
    """
    Displays the configuration of the processing pipeline.

    Parameters
    ----------
    args : argparse.Namespace
        The parsed command-line arguments.

    Notes
    -----
    This function generates a table displaying the configuration options specified via command-line arguments,
    providing a clear overview of the pipeline settings before execution.
    """
    table = Table(title="Processing Pipeline Config")
    table.add_column("Argument", style="cyan", no_wrap=True)
    table.add_column("Value", style="magenta")
    for arg in vars(args):
        table.add_row(arg.replace("_", " ").title(), str(getattr(args, arg)))
    console.print(table)


if __name__ == "__main__":
    main()



from ctdfjorder.constants.constants import *
from ctdfjorder.exceptions.exceptions import CTDError

from datetime import datetime, timedelta
import polars as pl

from os import path

# Column label mapping from castaway to internal
csvLabels_to_labelInternal: dict[str, str] = {
    "Pressure (Decibar)": SEA_PRESSURE_LABEL,
    "Depth (Meter)": DEPTH_LABEL,
    "Temperature (Celsius)": TEMPERATURE_LABEL,
    "Conductivity (MicroSiemens per Centimeter)": CONDUCTIVITY_LABEL,
    "Specific conductance (MicroSiemens per Centimeter)": SPECIFIC_CONDUCTIVITY_LABEL,
    "Salinity (Practical Salinity Scale)": SALINITY_LABEL,
    "Sound velocity (Meters per Second)": SPEED_OF_SOUND_LABEL,
    "Density (Kilograms per Cubic Meter)": DENSITY_LABEL,
}


def load_file_castaway(castaway_file_path):
    """
    Loads and processes a Castaway CTD file.

    Parameters
    ----------
    castaway_file_path : str
        The file path to the Castaway CTD file.

    Returns
    -------
    pl.DataFrame
        The processed Castaway CTD data.

    Raises
    ------
    CTDError
        If the Castaway CTD profile is empty or if no samples are found in the file, or if the start time is missing.
    """
    filename = path.basename(castaway_file_path)
    with open(castaway_file_path) as file:
        profile = pl.read_csv(file, comment_prefix="%", null_values=["#N/A", "null"])
    if profile.is_empty():
        raise CTDError(message=ERROR_NO_SAMPLES, filename=filename)
    if CASTAWAY_DATETIME_LABEL in profile.columns:
        profile = profile.with_columns(
            pl.col(CASTAWAY_DATETIME_LABEL)
            .str.to_datetime(
                format=TIME_FORMAT,
                time_zone=TIME_ZONE,
                time_unit=TIME_UNIT,
            )
            .cast(pl.Datetime)
            .alias(CASTAWAY_DATETIME_LABEL)
        )
        start_time = profile.select(pl.col(CASTAWAY_DATETIME_LABEL).first()).item()
    else:
        start_time = extract_utc_cast_time(castaway_file_path)
    if type(start_time) == type(None):
        raise CTDError(filename=filename, message=ERROR_CASTAWAY_START_TIME)
    timestamps = [
        start_time + timedelta(milliseconds=200 * i) for i in range(profile.height)
    ]
    profile = profile.with_columns(
        pl.Series(timestamps)
        .dt.convert_time_zone(TIME_ZONE)
        .dt.cast_time_unit(TIME_UNIT)
        .alias(TIMESTAMP_LABEL)
    )
    for header, maps_to in csvLabels_to_labelInternal.items():
        if header in profile.columns:
            profile = profile.rename({header: maps_to})
    profile = profile.drop(CASTAWAY_FILE_ID_LABEL, strict=False).with_columns(
        (pl.col(SEA_PRESSURE_LABEL) + 10.1325).alias(PRESSURE_LABEL),
        pl.lit(0).alias(PROFILE_ID_LABEL),
        pl.lit(filename).alias(FILENAME_LABEL),
    )
    if LATITUDE_LABEL not in profile.collect_schema().names():
        lat, long = extract_lat_long_castaway(castaway_file_path)
        profile = profile.with_columns(
            pl.lit(lat).cast(pl.Float64).alias(LATITUDE_LABEL),
            pl.lit(long).cast(pl.Float64).alias(LONGITUDE_LABEL),
        )
    data = profile
    return data


def extract_utc_cast_time(ctd_file_path):
    """
    Extracts the UTC cast time from a Castaway file and converts it to ISO 8601 format.

    Parameters
    ----------
    ctd_file_path : str
        The file path of the Castaway file to extract the time from.

    Returns
    -------
    datetime.datetime | None
        The cast time (UTC) of the Castaway file in ISO 8601 format, or None if not found.
    """
    # Initialize variable to store UTC cast time
    cast_time_utc = None
    # Open the file and read line by line
    with open(ctd_file_path, "r") as file:
        for line in file:
            if line.startswith("% Cast time (UTC)"):
                # Extract the UTC cast time after the comma
                parts = line.strip().split(",")
                if len(parts) > 1:
                    cast_time_str = parts[1].strip()  # Take only the datetime part
                    # Convert the datetime string to ISO format if possible
                    cast_time_utc = datetime.strptime(
                        cast_time_str, "%Y-%m-%d %H:%M:%S"
                    )
                    break  # Stop reading once the timestamp is found

    return cast_time_utc


def extract_lat_long_castaway(ctd_file_path):
    """
    Extracts the start latitude and longitude from a Castaway file.

    Parameters
    ----------
    ctd_file_path : str
        The file path to the Castaway CTD file.

    Returns
    -------
    tuple
        A tuple containing the latitude and longitude as strings, or None if not found.
    """
    latitude = None
    longitude = None
    # Open the file and read line by line
    with open(ctd_file_path, "r") as file:
        for line in file:
            if line.startswith("% Start latitude"):
                # Assume the format is '% description, latitude'
                parts = line.strip().split(",")
                if len(parts) > 1:
                    latitude = parts[1].strip()
                    if latitude == "":
                        latitude = None
            if line.startswith("% Start longitude"):
                # Assume the format is '% description, longitude'
                parts = line.strip().split(",")
                if len(parts) > 1:
                    longitude = parts[1].strip()
                    if longitude == "":
                        longitude = None
    return latitude, longitude



"""
Constants used throughout the CTDFjorder package.

This module contains various constant values used for labeling columns,
time formatting, error messages, warning messages, and other configuration
settings within the CTDFjorder package.
"""

from typing import Literal

# Column labels for internal use
TIMESTAMP_LABEL: str = "timestamp"
"""str: Label for timestamp column."""
YEAR_LABEL: str = "year"
"""str: Label for year column."""
MONTH_LABEL: str = "month"
"""str: Label for month column."""
FILENAME_LABEL: str = "filename"
"""str: Label for filename column."""
CHLOROPHYLL_LABEL: str = "chlorophyll"
"""str: Label for chlorophyll column."""
TEMPERATURE_LABEL: str = "temperature"
"""str: Label for temperature column."""
SEA_PRESSURE_LABEL: str = "sea_pressure"
"""str: Label for sea pressure column."""
DEPTH_LABEL: str = "depth"
"""str: Label for depth column."""
SALINITY_LABEL: str = "salinity"
"""str: Label for salinity column."""
SPEED_OF_SOUND_LABEL: str = "speed_of_sound"
"""str: Label for speed of sound column."""
SPECIFIC_CONDUCTIVITY_LABEL: str = "specific_conductivity"
"""str: Label for specific conductivity column."""
CONDUCTIVITY_LABEL: str = "conductivity"
"""str: Label for conductivity column."""
PRESSURE_LABEL: str = "pressure"
"""str: Label for pressure column."""
SALINITY_ABS_LABEL: str = "salinity_abs"
"""str: Label for absolute salinity column."""
SURFACE_SALINITY_LABEL: str = "surface_salinity"
"""str: Label for surface salinity column."""
SURFACE_TEMPERATURE_LABEL: str = "surface_temperature"
"""str: Label for surface temperature column."""
SURFACE_DENSITY_LABEL: str = "surface_density"
"""str: Label for surface density column."""
MELTWATER_FRACTION_LABEL: str = "meltwater_fraction"
"""str: Label for meltwater fraction column."""
DENSITY_LABEL: str = "density"
"""str: Label for density column."""
POTENTIAL_DENSITY_LABEL: str = "potential_density"
"""str: Label for potential density column."""
BV_LABEL: str = "brunt_vaisala_frequency_squared"
"""str: Label for Brunt-Väisälä frequency squared column."""
P_MID_LABEL: str = "p_mid"
"""str: Label for mid pressure used for Brunt-Väisälä calculation column."""
SECCHI_DEPTH_LABEL: str = "secchi_depth"
"""str: Label for Secchi depth column."""
LATITUDE_LABEL: str = "latitude"
"""str: Label for latitude column."""
LONGITUDE_LABEL: str = "longitude"
"""str: Label for longitude column."""
UNIQUE_ID_LABEL: str = "unique_id"
"""str: Label for unique ID column."""
PROFILE_ID_LABEL: str = "profile_id"
"""str: Label for profile ID column."""
SITE_NAME_LABEL: str = "site_name"
"""str: Label for site name column."""
SITE_ID_LABEL: str = "site_id"
"""str: Label for the site id column."""
LIST_LABELS: list[str] = [
    TIMESTAMP_LABEL,
    YEAR_LABEL,
    MONTH_LABEL,
    FILENAME_LABEL,
    CHLOROPHYLL_LABEL,
    TEMPERATURE_LABEL,
    SEA_PRESSURE_LABEL,
    DEPTH_LABEL,
    SALINITY_LABEL,
    SPEED_OF_SOUND_LABEL,
    SPECIFIC_CONDUCTIVITY_LABEL,
    CONDUCTIVITY_LABEL,
    PRESSURE_LABEL,
    SALINITY_ABS_LABEL,
    SURFACE_SALINITY_LABEL,
    SURFACE_TEMPERATURE_LABEL,
    SURFACE_DENSITY_LABEL,
    MELTWATER_FRACTION_LABEL,
    DENSITY_LABEL,
    POTENTIAL_DENSITY_LABEL,
    BV_LABEL,
    P_MID_LABEL,
    SECCHI_DEPTH_LABEL,
    LATITUDE_LABEL,
    LONGITUDE_LABEL,
    UNIQUE_ID_LABEL,
    PROFILE_ID_LABEL,
    SITE_ID_LABEL,
    SITE_NAME_LABEL
]
"""list[str]: List of all internal column labels."""

# Export labels
EXPORT_TIMESTAMP_LABEL = "timestamp"
"""str: Export label for timestamp."""
EXPORT_YEAR_LABEL = "Year"
"""str: Export label for year."""
EXPORT_MONTH_LABEL = "Month"
"""str: Export label for month."""
EXPORT_TEMPERATURE_LABEL = "Temperature_(°C)"
"""str: Export label for temperature."""
EXPORT_PRESSURE_LABEL = "Pressure_(dbar)"
"""str: Export label for pressure."""
EXPORT_DEPTH_LABEL = "Depth_(m)"
"""str: Export label for depth."""
EXPORT_SEA_PRESSURE_LABEL = "Sea_Pressure_(dbar)"
"""str: Export label for sea pressure."""
EXPORT_CHLOROPHYLL_LABEL = "Chlorophyll_a_(µg/l)"
"""str: Export label for chlorophyll-a."""
EXPORT_SALINITY_LABEL = "Salinity_(PSU)"
"""str: Export label for salinity."""
EXPORT_SPECIFIC_CONDUCTIVITY_LABEL = "Specific_Conductivity_(µS/cm)"
"""str: Export label for specific conductivity."""
EXPORT_CONDUCTIVITY_LABEL = "Conductivity_(mS/cm)"
"""str: Export label for conductivity."""
EXPORT_DENSITY_LABEL = "Density_(kg/m^3)"
"""str: Export label for density."""
EXPORT_POTENTIAL_DENSITY_LABEL = "Potential_Density_(kg/m^3)"
"""str: Export label for potential density."""
EXPORT_SALINITY_ABS_LABEL = "Absolute_Salinity_(g/kg)"
"""str: Export label for absolute salinity."""
EXPORT_SURFACE_DENSITY_LABEL = "Mean_Surface_Density_(kg/m^3)"
"""str: Export label for mean surface density."""
EXPORT_SPEED_OF_SOUND_LABEL = "Speed_of_Sound_(m/s)"
"""str: Export label for speed of sound."""
EXPORT_SURFACE_SALINITY_LABEL = "Surface_Salinity_(PSU)"
"""str: Export label for surface salinity."""
EXPORT_SURFACE_TEMPERATURE_LABEL = "Surface_Temperature_(°C)"
"""str: Export label for surface temperature."""
EXPORT_MELTWATER_FRACTION_LABEL = "Meltwater_Fraction_(%)"
"""str: Export label for meltwater fraction."""
EXPORT_LONGITUDE_LABEL = "longitude"
"""str: Export label for longitude."""
EXPORT_LATITUDE_LABEL = "latitude"
"""str: Export label for latitude."""
EXPORT_FILENAME_LABEL = "filename"
"""str: Export label for filename."""
EXPORT_PROFILE_ID_LABEL = "Profile_ID"
"""str: Export label for profile ID."""
EXPORT_UNIQUE_ID_LABEL = "Unique_ID"
"""str: Export label for unique ID."""
EXPORT_BV_LABEL = "Brunt_Vaisala_Frequency_Squared"
"""str: Export label for Brunt-Väisälä frequency squared."""
EXPORT_P_MID_LABEL = "Mid_Pressure_Used_For_BV_Calc"
"""str: Export label for mid pressure used for Brunt-Väisälä calculation."""
EXPORT_SECCHI_DEPTH_LABEL = "Secchi_Depth_(m)"
"""str: Export label for Secchi depth."""
EXPORT_SITE_NAME_LABEL = "Site_Name"
"""str: Export label for site name."""
EXPORT_SITE_ID_LABEL = "Site_ID"
"""str: Export label for site id."""

# Time string constants
TIME_ZONE: str = "UTC"
"""str: Time zone used for timestamps."""
TIME_UNIT: Literal["ns", "us", "ms"] = "ns"
"""Literal["ns", "us", "ms"]: Time unit for timestamps."""
TIME_FORMAT: str = "%Y-%m-%d %H:%M:%S.%f"
"""str: Format for timestamps."""

# Error messages
ERROR_NO_SAMPLES: str = "No samples in file"
"""str: Error message for no samples in file."""
ERROR_NO_LOCATION: str = "No location could be found"
"""str: Error message for no location found."""
ERROR_DENSITY_CALCULATION: str = "Could not calculate density on this dataset"
"""str: Error message for density calculation failure."""
ERROR_SALINITY_ABS: str = "Could not calculate salinity absolute on this dataset"
"""str: Error message for salinity absolute calculation failure."""
ERROR_NO_MASTER_SHEET: str = (
    "No mastersheet provided, could not update the file's missing location data"
)
"""str: Error message for missing master sheet."""
ERROR_RSK_CORRUPT: str = "Ruskin file is corrupted and could not be read"
"""str: Error message for corrupted Ruskin file."""
ERROR_LOCATION_DATA_INVALID: str = (
    "Location data invalid, probably due to malformed master sheet data"
)
"""str: Error message for invalid location data."""
ERROR_NO_TIMESTAMP_IN_FILE: str = "No timestamp in file, could not get location"
"""str: Error message for missing timestamp in file."""
ERROR_NO_TIMESTAMP_IN_MASTER_SHEET: str = (
    "No timestamp in master sheet, could not get location"
)
"""str: Error message for missing timestamp in master sheet."""
ERROR_MLD_DEPTH_RANGE: str = "Insufficient depth range to calculate MLD"
"""str: Error message for insufficient depth range to calculate MLD."""
ERROR_GRU_INSUFFICIENT_DATA: str = "Not enough values to run the GRU on this data"
"""str: Error message for insufficient data to run the GRU."""
ERROR_CASTAWAY_START_TIME: str = "Castaway file has no time data"
"""str: Error message for missing time data in Castaway file."""
ERROR_CTD_FILENAME_ENDING: str = "CTD filename must end in '.rsk' or '.csv'"
"""str: Error message for invalid CTD filename ending."""

# Warning messages
WARNING_DROPPED_PROFILE: str = "No samples in profile number "
"""str: Warning message for dropped profile due to no samples."""
WARNING_CTD_SURFACE_MEASUREMENT: str = (
    "First measurment lies below 'end'parameter, cannot compute surface measurements"
)
"""str: Warning message for invalid surface measurement."""
WARNING_FILE_LACKS_LOCATION: str = "File lacks native location data"
"""str: Warning message for file lacking native location data."""

# Filename constants
FILENAME_GPS_ENDING: str = "_gps"
"""str: Suffix for GPS filename."""
FILENAME_CM_ENDING: str = "cm"
"""str: Suffix for CM filename."""
RSK_FILE_MARKER: str = ".rsk"
"""str: Marker for RSK file."""
CASTAWAY_FILE_MARKER: str = ".csv"
"""str: Marker for Castaway file."""

# Castaway column labels
CASTAWAY_DATETIME_LABEL: str = "datetime_utc"
"""str: Column label for datetime in Castaway files."""
CASTAWAY_FILE_ID_LABEL: str = "file_id"
"""str: Column label for file ID in Castaway files."""

# Concatenation parameters
CONCAT_HOW: Literal["diagonal_relaxed"] = "diagonal_relaxed"
"""Literal["diagonal_relaxed"]: Concatenation method for merging data."""

# Sea pressure to pressure difference
SEA_PRESSURE_TO_PRESSURE_DIFF: float = 10.1325
"""float: Conversion factor from sea pressure to pressure difference."""

# Library logger name
LIB_LOGGER_NAME = "ctdfjorder"
"""str: Logger name for the CTDFjorder library."""

# Default output file
DEFAULT_OUTPUT_FILE = "ctdfjorder_data.csv"
"""str: Name of the default output file."""


import polars as pl
from os import path

import ctdfjorder.exceptions.exceptions
from ctdfjorder.exceptions.exceptions import (
    CTDError,
    raise_warning_improbable_match,
)
from ctdfjorder.constants.constants import *
from ctdfjorder.scardb.scardb import generate_sites_database
from ctdfjorder.dataclasses.dataclasses import SitesDatabase, Metadata

from typing import Literal


class MasterSheet:
    """
    Represents a master sheet for research sample data, allowing for the integration and cross-checking of sample data.

    Attributes
    ----------
    TIME_UNIT : Literal["ns", "us", "ms"], default "ns"
        The time unit for datetime operations.
    TIME_ZONE : str, default "UTC"
        The time zone for datetime operations.
    data : polars.DataFrame
        The data from the master sheet.
    secchi_depth_label : str
        The label for secchi depth in the master sheet.
    latitude_label : str
        The label for latitude in the master sheet.
    longitude_label : str
        The label for longitude in the master sheet.
    date_utc_label : str
        The label for UTC date in the master sheet.
    time_utc_label : str
        The label for UTC time in the master sheet.
    site_names_label : str
        The label for site names in the master sheet.
    site_names_short_label : str
        The label for short site names in the master sheet.
    with_crosschecked_site_names : bool
        Indicates whether site names should be cross-checked with a database.
    site_names_db : SitesDatabase
        The database of site names for cross-checking.

    Parameters
    ----------
    master_sheet_path : str
        The file path to the master sheet.
    secchi_depth_label : str, default "secchi depth"
        The label for secchi depth in the master sheet.
    latitude_label : str, default "latitude"
        The label for latitude in the master sheet.
    longitude_label : str, default "longitude"
        The label for longitude in the master sheet.
    datetime_utc_label : str, default "date/time (ISO)"
        The label for UTC date and time in the master sheet.
    unique_id_label : str, default "UNIQUE ID CODE "
        The label for unique ID in the master sheet.
    site_names_label : str, default "location"
        The label for site names in the master sheet.
    site_names_short_label : str, default "loc id"
        The label for short site names in the master sheet.
    with_crosschecked_site_names : bool, default False
        Indicates whether site names should be cross-checked with a database.

    Raises
    ------
    IOError
        If the master sheet file type is not supported or if the data cannot be read.
    CTDError
        If there are issues with the data, such as missing or invalid values.

    """
    TIME_UNIT: Literal["ns", "us", "ms"] = "ns"
    TIME_ZONE: str = "UTC"
    data = None
    # Column labels of master sheet
    secchi_depth_label: str
    latitude_label: str
    longitude_label: str
    date_utc_label: str
    time_utc_label: str
    filename_label: str
    site_names_label: str
    site_names_short_label: str
    with_crosschecked_site_names: bool
    site_names_db: SitesDatabase

    def __init__(
        self,
        master_sheet_path: str,
        secchi_depth_label: str = "secchi depth",
        latitude_label: str = "nominal latitude",
        longitude_label: str = "nominal longitude",
        datetime_utc_label: str = "date/time (ISO)",
        unique_id_label: str = "UNIQUE ID CODE ",
        filename_label: str = "CTD cast file name",
        site_names_label: str = "location",
        site_names_short_label: str = "loc id",
        with_crosschecked_site_names: bool = False,
    ):
        self.secchi_depth_label: str = secchi_depth_label
        self.latitude_label: str = latitude_label
        self.longitude_label: str = longitude_label
        self.datetime_utc_label: str = datetime_utc_label
        self.unique_id_label: str = unique_id_label
        self.site_names_label: str = site_names_label
        self.site_names_short_label: str = site_names_short_label
        self.filename_label: str = filename_label
        list_of_cols = [
            self.secchi_depth_label,
            self.latitude_label,
            self.longitude_label,
            self.datetime_utc_label,
            self.unique_id_label,
            self.site_names_label,
            self.site_names_short_label,
            self.filename_label
        ]
        self.with_crosschecked_site_names: bool = with_crosschecked_site_names

        # Generating polars dataframe representation of the master sheet
        df = None
        null_values = [
            "-999",
            "NA",
            "#N/A",
            "",
            "2022-10-29 -999",
            "14-11-2022 11:50",
            "14-11-2022 12:20",
            "2022-11-28 -999",
            "28-11-2022 13:00",
            "23-12-2022 20:30",
            "16-1-2023 11:37",
            "19-1-2023 13:23",
            "2023-01-22 -999",
            "17-2-2023 12:01",
            "17-2-2023 17:10",
            "18-2-2023 17:05",
            "19-2-2023 12:03",
            "20-2-2023 12:05",
            "20-2-2023 22:00",
            "20-2-2023 16:00",
            "22-2-2023 11:30",
            "22-2-2023 18:30",
            "24-2-2023 18:00",
            "25-2-2023 17:00",
            "26-2-2023 11:28",
            "27-2-2023 11:06",
            "27-2-2023 18:00",
            "28-Jan--22 18:30",
            "OCt-NOV ",
            " ",
        ]
        if ".xlsx" in path.basename(master_sheet_path):
            df = pl.read_excel(
                master_sheet_path,
                engine="calamine"
            )
            df = df.select(list_of_cols)
            df = df.with_columns(pl.col(self.datetime_utc_label).replace(old=list_of_cols, new=None))
            df = df.with_columns(pl.col(self.datetime_utc_label).str.to_datetime(format='%Y-%m-%dT%H:%M:%S.%f%z',
                                                                            strict=False,
                                                                                 exact=False)
                                 .cast(pl.Datetime)
                                 .dt.cast_time_unit(TIME_UNIT).dt.replace_time_zone(TIME_ZONE).alias(self.datetime_utc_label))
            df = df.drop_nulls(self.datetime_utc_label)
            if df.is_empty():
                raise ctdfjorder.exceptions.exceptions.Critical(f"Could not read mastersheet data from {master_sheet_path}."
                                                                    f" If on mac download your mastersheet as a csv not an xlsx.")

        if ".csv" in path.basename(master_sheet_path):
            df = pl.read_csv(
                master_sheet_path,
                columns=list_of_cols,
                try_parse_dates=True,
                rechunk=True,
                null_values=null_values,
            )
            try:
                df.select(pl.col(self.datetime_utc_label).dt.time())
            except pl.exceptions.SchemaError:
                raise ctdfjorder.exceptions.exceptions.Critical(f"Could not read mastersheet data from {master_sheet_path}."
                                                                f" If on mac download your mastersheet as a csv not an xlsx.")
        if type(df) is type(None):
            raise IOError(f"Invalid master sheet filetype. {master_sheet_path} not an xlsx or csv file.")
        self.data = df.drop_nulls(self.datetime_utc_label)
        self.data = self.data.with_columns(
            pl.col(latitude_label).cast(pl.Float64).alias(latitude_label),
            pl.col(longitude_label).cast(pl.Float64).alias(longitude_label),
        )
        if self.with_crosschecked_site_names:
            sites_in_master_sheet = (
                df.select(pl.col(site_names_label)).to_series().to_list()
            )
            self.site_names_db = generate_sites_database(sites_in_master_sheet)

    def find_match(
        self,
        profile: pl.DataFrame,
    ) -> Metadata:
        """
        Locates the row in the master sheet with a filename value that matches the profile parameter.
        Returns the latitude, longitude, unique id, and secchi depth from that row.
        based on the closest match.

        Parameters
        ----------
        profile : pl.DataFrame
            Profile to match to master sheet.

        Returns
        -------
        Metadata
            An object containing the estimated latitude, longitude, unique id, and secchi depth.

        Raises
        ------
        CTDError
            When there is no timestamp data in the master sheet and/or CTD file.
        Warning
            When the guessed unique ID or latitude is improbable or inconsistent.
        """
        filename = profile.select(pl.first(FILENAME_LABEL)).item()
        closest_row_overall = self.data.filter(pl.col(self.filename_label).str.contains(filename))
        if closest_row_overall.height < 1:
            raise CTDError(message="No unique ID's associated with this cast", filename=filename)
        if closest_row_overall.height > 1:
            raise CTDError(message="Multiple unique ID's associated with this cast", filename=filename)
        latitude = closest_row_overall.item(row=0, column=self.latitude_label)
        longitude = closest_row_overall.item(row=0, column=self.longitude_label)
        unique_id = closest_row_overall.select(pl.col(self.unique_id_label)).item(row=0, column=0)
        secchi_depth = closest_row_overall.select(
            pl.col(self.secchi_depth_label).cast(pl.Float32, strict=False).first()
        ).item(row=0, column=0)
        site_name = closest_row_overall.select(
            pl.col(self.site_names_label).cast(pl.String).first()
        ).item(row=0, column=0)
        site_id = closest_row_overall.select(
            pl.col(self.site_names_short_label).cast(pl.String).first()
        ).item(row=0, column=0)
        if not isinstance(latitude, float):
            raise_warning_improbable_match(
                f"Latitude: {latitude} is not float", filename=filename
            )
        if self.with_crosschecked_site_names:
            latitude_from_profile = profile.select(
                pl.col(LATITUDE_LABEL).first()
            ).item()
            longitude_from_profile = profile.select(
                pl.col(LONGITUDE_LABEL).first()
            ).item()
            if latitude_from_profile is not None and longitude_from_profile is not None:
                site_name_of_sample_master_sheet = closest_row_overall.select(
                    (pl.col(self.site_names_label))
                ).item()
                for site in self.site_names_db.sites:
                    if site_name_of_sample_master_sheet == site.name:
                        if (
                            not abs(site.latitude - latitude_from_profile) < 0.2
                            or not abs(site.longitude - longitude_from_profile) < 0.2
                        ):
                            message = (
                                f"Matched to Unique ID '{unique_id}' but crosschecking "
                                f"site location from SCAR with location from file yields "
                                f"inconsistent results."
                            )
                            raise CTDError(
                                message=message, filename=filename
                            )
        return Metadata(
            latitude=latitude,
            longitude=longitude,
            unique_id=unique_id,
            secchi_depth=secchi_depth,
            site_name=site_name,
            site_id=site_id
        )



# List of site names
import os
from ctdfjorder.dataclasses.dataclasses import SitesDatabase, ResearchSite
import polars as pl
from ctdfjorder.exceptions.exceptions import raise_warning_site_location


def generate_sites_database(site_names: list[str]) -> SitesDatabase:
    """
    Generates a database of research sites based on given site names, using data from various SCAR (Scientific Committee on Antarctic Research) datasets.

    Parameters
    ----------
    site_names : list[str]
        A list of site names to search for in the SCAR datasets.

    Returns
    -------
    SitesDatabase
        An object containing a list of ResearchSite objects for each matching site found in the datasets.

    Notes
    -----
    The function attempts to find exact matches for the provided site names in the SCAR datasets. If no exact match is found,
    it calculates the Levenshtein similarity between the provided site name and the names in the datasets to suggest potential matches.

    Examples
    --------
    Generate a database with a list of site names:

    >>> site_names = ['Site A', 'Site B', 'Site C']
    >>> database = generate_sites_database(site_names)
    >>> print(database)
    SitesDatabase with 3 sites

    Details
    -------
    - The function uses Levenshtein distance to compute similarities between site names.
    - It reads data from SCAR datasets stored in Parquet files.
    - If an exact match is found, the corresponding site information is added to the `SitesDatabase`.
    - If no exact match is found, it suggests potential matches with the highest similarity score.
    - Raises a warning if no similar site name is found in any of the datasets.

    Raises
    ------
    Warning
        If a provided site name does not match exactly or similarly with any site name in the SCAR datasets.

    """
    def levenshtein_distance(s1, s2):
        s1 = str.lower(s1)
        s2 = str.lower(s2)
        m, n = len(s1), len(s2)
        dp = [[0] * (n + 1) for _ in range(m + 1)]

        # Initialize the matrix
        for i in range(m + 1):
            dp[i][0] = i
        for j in range(n + 1):
            dp[0][j] = j

        # Compute the Levenshtein distance
        for i in range(1, m + 1):
            for j in range(1, n + 1):
                if s1[i - 1] == s2[j - 1]:
                    cost = 0
                else:
                    cost = 1
                dp[i][j] = min(
                    dp[i - 1][j] + 1,  # Deletion
                    dp[i][j - 1] + 1,  # Insertion
                    dp[i - 1][j - 1] + cost,
                )  # Substitution

        return dp[m][n]

    def levenshtein_similarity(s1, s2):
        if type(s1) is type(None):
            return 0.0
        if type(s2) is type(None):
            return 0.0
        max_len = max(len(s1), len(s2))
        if max_len == 0:
            return 1.0
        distance = levenshtein_distance(s1, s2)
        return float((max_len - distance) / max_len)

    def get_parquet_file_path(filename: str):
        # Get the absolute path of the Parquet file
        base_dir = os.path.dirname(__file__)
        data_dir = os.path.join(base_dir, "sitenamesdb")
        file_path = os.path.join(data_dir, filename)
        return file_path

    site_names_SCAR_USA = pl.read_parquet(get_parquet_file_path("USA.parquet"))
    site_names_SCAR_UK = pl.read_parquet(get_parquet_file_path("UK.parquet"))
    site_names_SCAR_Argentina = pl.read_parquet(
        get_parquet_file_path("Argentina.parquet")
    )
    site_names_SCAR_Chile = pl.read_parquet(get_parquet_file_path("Chile.parquet"))

    priority_dfs = [
        site_names_SCAR_USA,
        site_names_SCAR_UK,
        site_names_SCAR_Argentina,
        site_names_SCAR_Chile,
    ]
    sites_found_list: list[ResearchSite] = []
    for site_name in site_names:
        found = False
        for df in priority_dfs:
            matching_rows = df.filter(pl.col("place_name_mapping") == site_name)
            if not matching_rows.is_empty():
                site_name = matching_rows.select(
                    pl.col("place_name_mapping").first()
                ).item()
                site_lat = matching_rows.select(pl.col("latitude").first()).item()
                site_long = matching_rows.select(pl.col("longitude").first()).item()
                site_narrative = matching_rows.select(
                    pl.col("narrative").first()
                ).item()
                new_site = ResearchSite(
                    name=site_name,
                    short_name=None,
                    latitude=site_lat,
                    longitude=site_long,
                    narrative=site_narrative,
                )
                sites_found_list.append(new_site)
                found = True
                break
        if not found:
            name = pl.col("place_name_mapping")
            sim = "similarity"
            sim_col = pl.col("similarity")
            SCAR_with_similarity_USA = site_names_SCAR_USA.select(
                name.map_elements(lambda x: levenshtein_similarity(x, site_name))
            )
            SCAR_with_similarity_UK = site_names_SCAR_UK.select(
                name.map_elements(lambda x: levenshtein_similarity(x, site_name))
            )
            SCAR_with_similarity_Argentina = site_names_SCAR_Argentina.select(
                name.map_elements(lambda x: levenshtein_similarity(x, site_name))
            )
            SCAR_with_similarity_Chile = site_names_SCAR_Chile.select(
                name.map_elements(lambda x: levenshtein_similarity(x, site_name))
            )
            sim_USA = site_names_SCAR_USA.with_columns(
                SCAR_with_similarity_USA.to_series().alias(sim)
            )
            sim_UK = site_names_SCAR_UK.with_columns(
                SCAR_with_similarity_UK.to_series().alias(sim)
            )
            sim_Arg = site_names_SCAR_Argentina.with_columns(
                SCAR_with_similarity_Argentina.to_series().alias(sim)
            )
            sim_Chile = site_names_SCAR_Chile.with_columns(
                SCAR_with_similarity_Chile.to_series().alias(sim)
            )
            closest_USA = sim_USA.filter(sim_col == sim_col.max()).limit(2)
            closest_UK = sim_UK.filter(sim_col == sim_col.max()).limit(2)
            closest_Arg = sim_Arg.filter(sim_col == sim_col.max()).limit(2)
            closest_Chile = sim_Chile.filter(sim_col == sim_col.max()).limit(2)
            sn_USA = closest_USA.select(name).to_series().to_list()
            sn_UK = closest_UK.select(name).to_series().to_list()
            sn_Arg = closest_Arg.select(name).to_series().to_list()
            sn_Chile = closest_Chile.select(name).to_series().to_list()
            ss_USA = closest_USA.select(sim_col).to_series().to_list()
            ss_UK = closest_UK.select(sim_col).to_series().to_list()
            ss_Arg = closest_Arg.select(sim_col).to_series().to_list()
            ss_Chile = closest_Chile.select(sim_col).to_series().to_list()
            if 0.0 in ss_USA:
                sn_USA = None
            if 0.0 in ss_UK:
                sn_UK = None
            if 0.0 in ss_Arg:
                sn_UK = None
            if 0.0 in ss_Chile:
                sn_UK = None
            raise_warning_site_location(
                message=f"Site name '{site_name}' may be one of US: {sn_USA} UK: {sn_UK} AR: {sn_Arg} CH: {sn_Chile} "
            )
    return SitesDatabase(sites_found_list)



from matplotlib import pyplot as plt
import numpy as np
import os
from ctdfjorder.constants.constants import *
from statsmodels.api import nonparametric
import polars as pl
from dash import dcc, html, dash
from dash.dependencies import Input, Output, State
import plotly.express as px
import logging
from flask import Flask
import re


def plot_map(df: pl.DataFrame, mapbox_access_token):
    """
    Generates an interactive map using Plotly and Dash to visualize CTD profiles.

    Parameters
    ----------
    df : pl.DataFrame
        The CTD data to plot on the map.
    mapbox_access_token : str
        The Mapbox access token for rendering the map.

    Notes
    -----
    This function creates a Dash web application with interactive controls for filtering the data by year,
    season, unique ID, latitude, longitude, and date range. The filtered data is displayed on a Mapbox map.
    """
    px.set_mapbox_access_token(mapbox_access_token)

    df = df.with_columns(((pl.col(EXPORT_MONTH_LABEL) % 12 + 3) // 3).alias("season"))
    lat_median = df.select(pl.col(EXPORT_LATITUDE_LABEL).median().first()).item()
    long_median = df.select(pl.col(EXPORT_LONGITUDE_LABEL).median().first()).item()
    pd_df = df.to_pandas()

    season_dict = {1: "Autumn", 2: "Winter", 3: "Spring", 4: "Summer"}

    server = Flask(__name__)
    app = dash.Dash(__name__, server=server)
    app.logger.setLevel(logging.ERROR)
    server.logger.setLevel(logging.ERROR)

    app.layout = html.Div(
        [
            html.Div(
                [
                    dcc.Dropdown(
                        id="year-dropdown",
                        options=[
                            {"label": str(year), "value": year}
                            for year in pd_df[EXPORT_YEAR_LABEL].unique()
                        ],
                        placeholder="Select a year",
                        multi=True,
                    ),
                    dcc.Dropdown(
                        id="season-dropdown",
                        options=[
                            {"label": "Winter", "value": 1},
                            {"label": "Spring", "value": 2},
                            {"label": "Summer", "value": 3},
                            {"label": "Autumn", "value": 4},
                        ],
                        placeholder="Select a season",
                        multi=True,
                    ),
                    dcc.Input(
                        id="unique-id-input",
                        type="text",
                        placeholder="Enter Unique ID or regex",
                        style={"width": "100%"},
                    ),
                    dcc.Input(
                        id="lat-range",
                        type="text",
                        placeholder="Enter latitude range (e.g. -60, -55)",
                        style={"width": "100%"},
                    ),
                    dcc.Input(
                        id="lon-range",
                        type="text",
                        placeholder="Enter longitude range (e.g. -70, -65)",
                        style={"width": "100%"},
                    ),
                    dcc.DatePickerRange(
                        id="date-picker-range",
                        start_date=pd_df["timestamp"].min(),
                        end_date=pd_df["timestamp"].max(),
                        display_format="YYYY-MM-DD",
                        style={"width": "100%"},
                    ),
                    html.Div(id="total-profiles", style={"padding": "10px 0"}),
                    html.Button("Download CSV", id="download-button"),
                    dcc.Download(id="download-dataframe-csv"),
                ],
                style={
                    "width": "20%",
                    "display": "inline-block",
                    "verticalAlign": "top",
                },
            ),
            html.Div(
                [
                    dcc.Graph(id="map", style={"height": "100vh", "width": "100vw"}),
                ],
                style={
                    "width": "80%",
                    "display": "inline-block",
                    "verticalAlign": "top",
                },
            ),
        ],
        style={"height": "100vh", "width": "100vw", "margin": 0, "padding": 0},
    )

    @app.callback(
        [Output("map", "figure"), Output("total-profiles", "children")],
        [
            Input("year-dropdown", "value"),
            Input("season-dropdown", "value"),
            Input("unique-id-input", "value"),
            Input("lat-range", "value"),
            Input("lon-range", "value"),
            Input("date-picker-range", "start_date"),
            Input("date-picker-range", "end_date"),
        ],
    )
    def update_map(
        selected_years,
        selected_seasons,
        unique_id_filter,
        lat_range,
        lon_range,
        start_date,
        end_date,
    ):
        filtered_df = pd_df.copy()
        if selected_years:
            filtered_df = filtered_df[
                filtered_df[EXPORT_YEAR_LABEL].isin(selected_years)
            ]
        if selected_seasons:
            filtered_df = filtered_df[filtered_df["season"].isin(selected_seasons)]
        if unique_id_filter:
            unique_id_list = unique_id_filter.split()
            try:
                filtered_df = filtered_df[
                    filtered_df["Unique_ID"].str.contains(
                        "|".join(unique_id_list), regex=True
                    )
                ]
            except re.error:
                pass
        if lat_range:
            try:
                lat_min, lat_max = map(float, lat_range.split(","))
                filtered_df = filtered_df[
                    (filtered_df["latitude"] >= lat_min)
                    & (filtered_df["latitude"] <= lat_max)
                ]
            except ValueError:
                pass
        if lon_range:
            try:
                lon_min, lon_max = map(float, lon_range.split(","))
                filtered_df = filtered_df[
                    (filtered_df["longitude"] >= lon_min)
                    & (filtered_df["longitude"] <= lon_max)
                ]
            except ValueError:
                pass
        if start_date and end_date:
            filtered_df = filtered_df[
                (filtered_df["timestamp"] >= start_date)
                & (filtered_df["timestamp"] <= end_date)
            ]

        total_profiles = len(filtered_df)
        try:
            lat_median = filtered_df[EXPORT_LATITUDE_LABEL].median()
            long_median = filtered_df[EXPORT_LONGITUDE_LABEL].median()
        except Exception as e:
            pass
        fig = px.scatter_mapbox(
            filtered_df,
            lat="latitude",
            lon="longitude",
            hover_name="Unique_ID",
            hover_data={
                "timestamp": True,
                "filename": True,
                "latitude": True,
                "longitude": True,
            },
            mapbox_style="dark",
            zoom=5,
            center={"lat": lat_median, "lon": long_median},
        )

        fig.update_layout(
            mapbox_accesstoken=mapbox_access_token, margin=dict(l=0, r=0, t=0, b=0)
        )
        return fig, f"Total Profiles: {total_profiles}"

    @app.callback(
        Output("download-dataframe-csv", "data"),
        [Input("download-button", "n_clicks")],
        [
            State("year-dropdown", "value"),
            State("season-dropdown", "value"),
            State("unique-id-input", "value"),
            State("lat-range", "value"),
            State("lon-range", "value"),
            State("date-picker-range", "start_date"),
            State("date-picker-range", "end_date"),
        ],
    )
    def download_csv(
        n_clicks,
        selected_years,
        selected_seasons,
        unique_id_filter,
        lat_range,
        lon_range,
        start_date,
        end_date,
    ):
        if n_clicks is None:
            raise dash.PreventUpdate

        filtered_df = pd_df.copy()
        if selected_years:
            filtered_df = filtered_df[
                filtered_df[EXPORT_YEAR_LABEL].isin(selected_years)
            ]
        if selected_seasons:
            filtered_df = filtered_df[filtered_df["season"].isin(selected_seasons)]
        if unique_id_filter:
            unique_id_list = unique_id_filter.split()
            try:
                filtered_df = filtered_df[
                    filtered_df["Unique_ID"].str.contains(
                        "|".join(unique_id_list), regex=True
                    )
                ]
            except re.error:
                pass
        if lat_range:
            try:
                lat_min, lat_max = map(float, lat_range.split(","))
                filtered_df = filtered_df[
                    (filtered_df["latitude"] >= lat_min)
                    & (filtered_df["latitude"] <= lat_max)
                ]
            except ValueError:
                pass
        if lon_range:
            try:
                lon_min, lon_max = map(float, lon_range.split(","))
                filtered_df = filtered_df[
                    (filtered_df["longitude"] >= lon_min)
                    & (filtered_df["longitude"] <= lon_max)
                ]
            except ValueError:
                pass
        if start_date and end_date:
            filtered_df = filtered_df[
                (filtered_df["timestamp"] >= start_date)
                & (filtered_df["timestamp"] <= end_date)
            ]

        return dcc.send_data_frame(filtered_df.to_csv, "selected_profiles.csv")

    import webbrowser
    from threading import Timer

    def open_browser():
        webbrowser.open_new("http://127.0.0.1:8050/")

    Timer(1, open_browser).start()
    app.run_server(debug=False)


def plot_depth_vs(
    df: pl.DataFrame, measurement: str, plot_folder: str, plot_type: str = "scatter"
):
    """
    Generates a plot of depth vs. specified measurement (salinity, density, temperature).

    Parameters
    ----------
    df : pl.DataFrame
        The CTD dataframe containing the data to plot.
    measurement : str
        The measurement to plot against depth. Options are 'salinity', 'density', 'potential_density', or 'temperature'.
    plot_folder : str
        The path to the folder where plots will be saved.
    plot_type : str, optional
        The type of plot to generate. Options are 'line' or 'scatter'. Defaults to 'scatter'.

    Notes
    -----
    - Adds horizontal lines indicating the mixed layer depth (MLD) if present.
    - Allows for both scatter and line plot types.
    - Saves the plot as an image file in the specified folder.
    """
    plt.rcParams.update({"font.size": 16})
    os.makedirs(plot_folder, exist_ok=True)
    for profile_id in (
        df.select(PROFILE_ID_LABEL).unique(keep="first").to_series().to_list()
    ):
        profile = df.filter(pl.col(PROFILE_ID_LABEL) == profile_id)
        filename = profile.select(pl.first(FILENAME_LABEL)).item()
        fig, ax1 = plt.subplots(figsize=(18, 18))
        ax1.invert_yaxis()
        ax1.set_ylim([profile.select(pl.col(DEPTH_LABEL)).max().item(), 0])
        color_map = {
            SALINITY_LABEL: "tab:blue",
            DENSITY_LABEL: "tab:red",
            POTENTIAL_DENSITY_LABEL: "tab:red",
            TEMPERATURE_LABEL: "tab:blue",
        }
        label_map = {
            SALINITY_LABEL: "Practical Salinity (PSU)",
            DENSITY_LABEL: "Density (kg/m^3)",
            POTENTIAL_DENSITY_LABEL: "Potential Density (kg/m^3)",
            TEMPERATURE_LABEL: "Temperature (°C)",
        }
        if plot_type == "line":
            lowess = nonparametric.lowess
            y, x = zip(
                *lowess(
                    profile.select(pl.col(f"{measurement}")).to_numpy(),
                    profile.select(pl.col(DEPTH_LABEL)).to_numpy(),
                    frac=0.1,
                )
            )
        else:
            x, y = (
                profile.select(pl.col(f"{measurement}")).to_numpy(),
                profile.select(pl.col(DEPTH_LABEL)).to_numpy(),
            )
        (
            ax1.plot(x, y, color=color_map[measurement], label=label_map[measurement])
            if plot_type == "line"
            else ax1.scatter(
                x, y, color=color_map[measurement], label=label_map[measurement]
            )
        )
        ax1.set_xlabel(label_map[measurement], color=color_map[measurement])
        ax1.set_ylabel("Depth (m)")
        ax1.tick_params(axis="x", labelcolor=color_map[measurement])
        mld = profile.select(pl.col(r"^.*MLD.*$").first()).item()
        if mld is not None and not np.isnan(mld):
            # Plot MLD line
            ax1.axhline(
                y=mld, color="green", linestyle="--", linewidth=2, label=f"{mld}"
            )
            ax1.text(
                0.95,
                mld,
                f"{mld}",
                va="center",
                ha="right",
                backgroundcolor="white",
                color="green",
                transform=ax1.get_yaxis_transform(),
            )
        plt.title(
            f"{filename} \n Profile {profile_id} \n Depth vs. {label_map[measurement]}\n MLD {mld}"
        )
        ax1.grid(True)
        ax1.legend(loc="upper center", bbox_to_anchor=(0.5, -0.15), ncol=3)
        plot_path = os.path.join(
            plot_folder,
            f"{filename}_{profile_id}_depth_{measurement}_{plot_type}_plot.png",
        )
        plt.savefig(plot_path)
        plt.close(fig)


def plot_original_data(salinity, depths, filename, plot_path):
    """
    Generates a scatter plot of original salinity vs. depth.

    Parameters
    ----------
    salinity : array-like
        The salinity data to plot.
    depths : array-like
        The corresponding depth data.
    filename : str
        The name of the file being plotted.
    plot_path : str
        The path to save the plot image.

    Returns
    -------
    tuple
        The x and y limits of the plot.

    Notes
    -----
    - Creates a scatter plot with salinity on the x-axis and depth on the y-axis (inverted).
    - Adds a title, labels, and grid to the plot.
    - Saves the plot as an image file and returns the plot limits.
    """
    plt.figure(figsize=(10, 6))
    plt.scatter(salinity, -depths, alpha=0.6)
    plt.xlabel("Salinity (PSU)")
    plt.ylabel("Depth (m)")
    plt.title(f"Original Salinity vs. Depth - {filename}")
    plt.grid(True)
    plt.savefig(plot_path)
    xlim = plt.xlim()
    ylim = plt.ylim()
    plt.close()
    return xlim, ylim


def plot_predicted_data(salinity, depths, xlim, ylim, filename, plot_path):
    """
    Generates a scatter plot of predicted salinity vs. depth.

    Parameters
    ----------
    salinity : array-like
        The predicted salinity data to plot.
    depths : array-like
        The corresponding depth data.
    xlim : tuple
        The x-axis limits from the original data plot.
    ylim : tuple
        The y-axis limits from the original data plot.
    filename : str
        The name of the file being plotted.
    plot_path : str
        The path to save the plot image.

    Notes
    -----
    - Creates a scatter plot with predicted salinity on the x-axis and depth on the y-axis (inverted).
    - Adds a title, labels, and grid to the plot.
    - Uses the same axis limits as the original data plot for comparison.
    - Saves the plot as an image file.
    """
    plt.figure(figsize=(10, 6))
    plt.scatter(salinity, -depths, alpha=0.6, color="red")
    title = f"Predicted Salinity vs. Depth - {filename}"
    plt.title(title)
    plt.xlabel("Salinity (PSU)")
    plt.ylabel("Depth (m)")
    plt.grid(True)
    plt.xlim((xlim[0], xlim[1]))
    plt.ylim((ylim[0], ylim[1]))
    plt.savefig(plot_path)
    plt.close()



from torch import optim

from ctdfjorder.constants.constants import *
from ctdfjorder.exceptions.exceptions import CTDError
from ctdfjorder.visualize.ctd_plot import plot_original_data, plot_predicted_data
import torch
from torch.utils.data import TensorDataset, DataLoader

import polars as pl
import numpy as np

from os import path
from os import getcwd

from sklearn.preprocessing import MinMaxScaler


class AI:
    """
    This class should not be imported, see :func:`ctdfjorder.CTD.clean` to use the methods here.

    """

    def clean_salinity_ai(profile: pl.DataFrame, profile_id: int) -> pl.DataFrame:
        r"""
        Cleans salinity using a GRU (Gated Recurrent Unit) machine learning model.

        Parameters
        ----------
        profile : pl.DataFrame
            Single profile of CTD (Conductivity, Temperature, Depth) data.
        profile_id : int
            Profile number, used for identification and logging.

        Returns
        -------
        pl.DataFrame
            CTD data with cleaned salinity values.

        Notes
        -----
        - This process bins the data every 0.5 dbar of pressure.
        - The GRU model uses a 16-width layer with a loss function designed to penalize decreasing salinity values with respect to pressure.
        - The penalization term is weighted to emphasize the importance of non-decreasing salinity with increasing pressure.

        GRU Model
        ---------
        The GRU model consists of:

        * An input layer that takes the binned CTD data.
        * A GRU layer with 16 units.
        * An output layer that predicts the salinity values.

        Loss Function
        -------------
        The custom loss function used in this model is the Mean Absolute Error (MAE) with an additional penalty term for salinity predictions that decrease with pressure. This is given by:

        .. math::
            L = \text{MAE}(y_{true}, y_{pred}) + \lambda \cdot \text{mean}(\text{penalties})

        where penalties are calculated as:

        .. math::
            \text{penalties} = \text{where}(\Delta s_{pred} < 0, \min(\Delta s_{pred}, 0), 0)

        Here, :math:`\Delta s_{pred}` represents the change in predicted salinity values, and :math:`\lambda` is the weighting factor for the penalty term.

        Examples
        --------
        >>> cleaned_data = AI.clean_salinity_ai(profile, profile_id=1)
        >>> cleaned_data.head()

        """

        class GRUModel(torch.nn.Module):
            def __init__(self, input_shape):
                super(GRUModel, self).__init__()
                self.gru = torch.nn.GRU(input_shape[1], 16, batch_first=True)
                self.output_layer = torch.nn.Linear(16, input_shape[1])

            def forward(self, x):
                gru_out, _ = self.gru(x)
                output = self.output_layer(gru_out)
                return output

        def loss_function(y_true, y_pred):
            """
            MAE loss with additional term to penalize salinity predictions that increase with pressure.
            Parameters
            ----------
            y_true
                True salinity tensor.
            y_pred
                Predicted salinity tensor.
            Returns
            -------
            torch.Tensor
                Loss value as a tensor.
            """
            # Assuming salinity is at index 0
            salinity_true = y_true[:, :, 0]
            salinity_pred = y_pred[:, :, 0]

            # Calculate differences between consecutive values
            delta_sal_pred = salinity_pred[:, 1:] - salinity_pred[:, :-1]

            # Penalize predictions where salinity decreases while pressure increases
            penalties = torch.where(
                delta_sal_pred < 0,
                -torch.min(
                    delta_sal_pred, torch.tensor(0.0, device=delta_sal_pred.device)
                ),
                torch.tensor(0.0, device=delta_sal_pred.device),
            )

            # Calculate mean absolute error
            mae = torch.mean(torch.abs(y_true - y_pred))

            # Add penalties
            total_loss = mae + 12.0 * torch.mean(
                penalties
            )  # Adjust weighting of penalty as needed
            return total_loss

        def build_gru(input_shape):
            model = GRUModel(input_shape)
            optimizer = optim.Adam(model.parameters(), lr=0.01)
            return model, optimizer

        def run_gru(data: pl.DataFrame, show_plots=True):
            """
            Runs the GRU.

            Parameters
            ----------
            data : DataFrame
                CTD dataframe
            show_plots : bool, default False
                True saves plots in working directory within the plots folder.
            Returns
            -------
            DataFrame
                CTD data with clean salinity values.

            Raises
            ------
            CTDError
                When there are not enough values to train on.

            """
            filename = data.select(pl.col(FILENAME_LABEL).first()).item()
            filtered_data = data.filter(pl.col(DEPTH_LABEL) > 1)
            filtered_data = filtered_data.with_columns(
                (pl.col(PRESSURE_LABEL) // 0.5 * 0.5).alias("pressure_bin")
            )
            # Define the desired columns and their aggregation functions
            column_agg_dict = {
                TEMPERATURE_LABEL: pl.mean(TEMPERATURE_LABEL),
                YEAR_LABEL: pl.first(YEAR_LABEL),
                MONTH_LABEL: pl.first(MONTH_LABEL),
                CHLOROPHYLL_LABEL: pl.mean(CHLOROPHYLL_LABEL),
                SEA_PRESSURE_LABEL: pl.mean(SEA_PRESSURE_LABEL),
                DEPTH_LABEL: pl.mean(DEPTH_LABEL),
                SALINITY_LABEL: pl.median(SALINITY_LABEL),
                SPEED_OF_SOUND_LABEL: pl.mean(SPEED_OF_SOUND_LABEL),
                SPECIFIC_CONDUCTIVITY_LABEL: pl.mean(SPECIFIC_CONDUCTIVITY_LABEL),
                CONDUCTIVITY_LABEL: pl.mean(CONDUCTIVITY_LABEL),
                DENSITY_LABEL: pl.mean(DENSITY_LABEL),
                POTENTIAL_DENSITY_LABEL: pl.mean(POTENTIAL_DENSITY_LABEL),
                SALINITY_ABS_LABEL: pl.mean(SALINITY_ABS_LABEL),
                TIMESTAMP_LABEL: pl.first(TIMESTAMP_LABEL),
                LONGITUDE_LABEL: pl.first(LONGITUDE_LABEL),
                LATITUDE_LABEL: pl.first(LATITUDE_LABEL),
                UNIQUE_ID_LABEL: pl.first(UNIQUE_ID_LABEL),
                FILENAME_LABEL: pl.first(FILENAME_LABEL),
                PROFILE_ID_LABEL: pl.first(PROFILE_ID_LABEL),
                SECCHI_DEPTH_LABEL: pl.first(SECCHI_DEPTH_LABEL),
                SITE_NAME_LABEL: pl.first(SITE_NAME_LABEL),
                SITE_ID_LABEL: pl.first(SITE_ID_LABEL),
            }
            available_columns = {
                col: agg_func
                for col, agg_func in column_agg_dict.items()
                if col in data.columns
            }
            data_binned = filtered_data.group_by(
                "pressure_bin", maintain_order=True
            ).agg(list(available_columns.values()))
            data_binned = data_binned.rename({"pressure_bin": PRESSURE_LABEL})
            if data_binned.limit(4).height < 2:
                raise CTDError(
                    message=ERROR_GRU_INSUFFICIENT_DATA,
                    filename=filename,
                )
            salinity = np.array(data_binned.select(pl.col(SALINITY_LABEL)).to_numpy())
            scaler = MinMaxScaler(feature_range=(-1, 1))
            scaler.fit(salinity)
            scaled_sequence = scaler.transform(salinity)
            scaled_seq = np.expand_dims(scaled_sequence, axis=0)
            min_pres = data_binned.select(pl.min(DEPTH_LABEL)).item()
            max_pres = data_binned.select(pl.max(DEPTH_LABEL)).item()
            pres_range = max_pres - min_pres
            epochs = int(pres_range * 12)
            input_shape = scaled_seq.shape[1:]
            model, optimizer = build_gru(input_shape)
            criterion = loss_function
            tensor_data = torch.tensor(scaled_seq, dtype=torch.float32)
            dataset = TensorDataset(tensor_data, tensor_data)
            data_loader = DataLoader(dataset, batch_size=4, shuffle=False)
            for epoch in range(epochs):
                model.train()
                for x_batch, y_batch in data_loader:
                    optimizer.zero_grad()  # Zero the gradients
                    y_pred = model(x_batch)  # Forward pass
                    loss = loss_function(y_batch, y_pred)  # Compute the loss
                    loss.backward()  # Backward pass
                    optimizer.step()  # Update the weights
            model.eval()
            with torch.no_grad():
                X_pred = model(tensor_data).numpy()
            predicted_seq = np.array(scaler.inverse_transform(X_pred[0])).flatten()
            if show_plots:
                xlim, ylim = plot_original_data(
                    data.select(SALINITY_LABEL).to_numpy(),
                    data.select(DEPTH_LABEL).to_numpy(),
                    filename + str(profile_id),
                    plot_path=path.join(
                        getcwd(), "ctdplots", f"{filename}_original.png"
                    ),
                )
                plot_predicted_data(
                    salinity=predicted_seq,
                    depths=data_binned.select(DEPTH_LABEL).to_numpy(),
                    filename=filename + str(profile_id),
                    xlim=xlim,
                    ylim=ylim,
                    plot_path=path.join(
                        getcwd(), "ctdplots", f"{filename}_predicted.png"
                    ),
                )
            data_binned = data_binned.with_columns(
                pl.Series(predicted_seq, dtype=pl.Float64).alias(SALINITY_LABEL)
            )
            return data_binned

        return run_gru(profile)
